/*
 * AlsaPlayer WMA plugin.
 *
 * Copyright (C) 2004 McMCC <mcmcc@mail.ru>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 *  $Id: wma_engine.c,v 1.0 2004/03/04 02:44:18 mcmcc Exp $
 *
 */ 

#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include "alsaplayer_error.h"
#include "input_plugin.h"
#include "reader.h"

#ifdef HAVE_AV_CONFIG_H
#undef HAVE_AV_CONFIG_H
#endif

#include "avcodec.h"
#include "avformat.h"

#define BLOCK_SIZE 1024

struct wma_local_data {
    AVFormatContext *ic;
    AVCodecContext *c;
    AVCodec *codec;
    AVPacket pkt;
    FifoBuffer f;
    uint8_t *inbuf_ptr;
    uint8_t *outbuf;
    int limit;
    int len;
    int size;
    int idx;
    int res;
    int player;
    int out_size;
    int seekpos;
    char path[FILENAME_MAX+1];
};		

static int wma_sample_rate(input_object *obj)
{
	struct wma_local_data *data;

	if (!obj)
		return 44100;
	data = (struct wma_local_data *)obj->local_data;
	if (data)
	    if(data->c->channels < 2)
		return data->c->sample_rate/2; /* mono */
	    else 
		return data->c->sample_rate;   /* stereo */
	return 44100;
}

static int wma_frame_seek(input_object *obj, int frame)
{
	struct wma_local_data *data;
	int64_t pos;

	if (!obj)
		return 0;
	data = (struct wma_local_data *)obj->local_data;
	if (data) {
		if (obj->flags & P_STREAMBASED)
			return 0;
		data->seekpos = (frame * BLOCK_SIZE ) / (wma_sample_rate(obj) * 2 * 2 / 100);
		return frame;
	}		
	return 0;
}

static int wma_frame_size(input_object *obj)
{
	if (!obj) {
		puts("No frame size!!!!");
		return 0;
	}
	return obj->frame_size;
}

static int wma_play_frame(input_object *obj, char *buf)
{
    struct wma_local_data *data;
    
    if (!obj || !obj->local_data)
	    return 0;
    data = (struct wma_local_data *)obj->local_data;

frame_new:    
    if(!data->player)
    {
	if(!data->limit)
	{
	    if (data->seekpos != -1)
	    {
    		av_seek_frame(data->ic, data->idx, data->seekpos * 10000LL);
    		data->seekpos = -1;
	    }
	    if (av_read_frame(data->ic, &data->pkt) < 0)
    		return 0;
	    
	    data->size = data->pkt.size;
	    data->inbuf_ptr = data->pkt.data;

	    if (data->size <= 0)
	    {
		data->limit = 0;
		data->player = 0;
        	return 0;
	    }
	}
	
	data->len = avcodec_decode_audio(data->c, (short*)data->outbuf, 
					 &data->out_size,
				         data->inbuf_ptr, 
					 data->size);
	/* skip bad frames or end */
    	if (data->len < 0)
	{
	    data->limit = 0;
	    data->player = 0;
    	    if (data->pkt.data)
        	av_free_packet(&data->pkt);
	    goto frame_new; 
	}

    	if (data->out_size <= 0)
	{
	    data->limit = 0;
	    data->player = 0;
    	    if (data->pkt.data)
        	av_free_packet(&data->pkt);
	    goto frame_new;
	}
	        
	fifo_init(&data->f, data->out_size*2);
	fifo_write(&data->f, data->outbuf, data->out_size, &data->f.wptr);
	data->res = 1;
    }

    data->player = 0;
    data->res = 1;
    
    memset(buf, 0, BLOCK_SIZE);
    if(!fifo_read(&data->f, buf, obj->frame_size, &data->f.rptr))
    {
        data->player = 1;
        data->res = 0;
    }
    
    if (data->seekpos != -1){
        data->limit = 0;
        data->res = 1;
        data->player = 0;
    }
    
    if(data->res)
    {
	data->limit = 0;
	data->res = 0;
	fifo_free(&data->f);    
        data->size -= data->len;
        data->inbuf_ptr += data->len;
        if (data->pkt.data)
            av_free_packet(&data->pkt);
	if(data->size > 0)
	    data->limit = 1;
	goto frame_new;
    }
    return 1;
}

static long wma_frame_to_sec(input_object *obj, int frame)
{
	struct wma_local_data *data;
	int sec;
	
	if (!obj || !obj->local_data)
		return 0;

	data = (struct wma_local_data *)obj->local_data;
	sec = (frame * BLOCK_SIZE ) / (wma_sample_rate(obj) * 2 * 2 / 100);
	return sec;
}

static int wma_nr_frames(input_object *obj)
{
	struct wma_local_data *data;
	int frames;
	
	if (!obj || !obj->local_data)
		return 0;
		
	data = (struct wma_local_data *)obj->local_data;
	if (!data)
		return 0;
	frames = ((data->ic->duration * 
		 (wma_sample_rate(obj) * 2 * 2 / 100))/BLOCK_SIZE)/10000LL;
	return frames;
}

int wma_stream_info(input_object *obj, stream_info *info)
{
	struct wma_local_data *data;	
	char y[16], n[4];

	if (!obj || !info)
		return 0;

	data = (struct wma_local_data *)obj->local_data;

	strncpy (info->path, data->path, sizeof (info->path));
	if (data) {
	    strcpy (info->title, data->path);
	    info->artist [0] = '\0';
	    info->album [0] = '\0';
	    info->genre [0] = '\0';
	    info->year [0] = '\0';
	    info->track [0] = '\0';
	    info->comment [0] = '\0';
	    if (data->ic->title[0] != '\0')
		strcpy(info->title, data->ic->title);
	    if (data->ic->author[0] != '\0')
		strcpy(info->artist, data->ic->author);
	    if (data->ic->album[0] != '\0')
		strcpy(info->album, data->ic->album);
	    if (data->ic->year != 0)
	    {
		sprintf(y,"%d", data->ic->year);
		strcpy(info->year, y);
	    }
	    if (data->ic->track != 0)
	    {
		sprintf(n,"%d", data->ic->track);
		strcpy(info->track, n);
	    }
	    if (data->ic->genre[0] != '\0')
		strcpy(info->genre, data->ic->genre);
	    if (data->ic->comment[0] != '\0')
		strcpy(info->comment, data->ic->comment);
	    if (data->c->sample_rate && data->c->bit_rate && data->codec->name) {
		sprintf(info->stream_type, "%s %dKHz %s %-3dkbit",
			data->codec->name,
			(int)(data->c->sample_rate / 1000), 
			data->c->channels == 1 ? "mono":"stereo",
			(int)(data->c->bit_rate / 1000));
	    } else {
		strcpy(info->stream_type, "Unknown WMA");
	    }	
	    info->status[0] = 0;
	}
	return 1;
}

static int wma_init(void) 
{
	avcodec_init();
	avcodec_register_all();
	av_register_all();
	return 1;
}		

static int wma_channels(input_object *obj)
{
	struct wma_local_data *data;

	if (!obj)
		return 2;
	data = (struct wma_local_data *)obj->local_data;
	if (data) 
	    return data->c->channels;
	return 2;
}

static float wma_can_handle(const char *path)
{
	char *ext;

	ext = strrchr(path, '.');

	if (ext)
		ext++;
	if (!ext)
		return 0.0;
	
	if (!strncasecmp(ext, "wma", 3)) {
		return 1.0;
	} else
		return 0.0;
}

static int wma_open(input_object *obj, const char *path)
{
    struct wma_local_data *data;

    if (!obj)
	return 0;

    obj->nr_channels = 2;
    obj->frame_size = BLOCK_SIZE;
    obj->local_data = malloc(sizeof(struct wma_local_data));
    if (!obj->local_data)
	return 0;
    data = (struct wma_local_data *)obj->local_data;
    if (!data) 
	return 0;

    obj->flags |= P_SEEK;
    obj->flags |= P_PERFECTSEEK;
    obj->flags |= P_FILEBASED;

    if(av_open_input_file(&data->ic, path, NULL, 0, NULL) < 0)
	return 0;

    for(data->idx = 0; data->idx < data->ic->nb_streams; data->idx++) {
        data->c = &data->ic->streams[data->idx]->codec;
        if(data->c->codec_type == CODEC_TYPE_AUDIO)
            break;
    }

    av_find_stream_info(data->ic);

    data->codec = avcodec_find_decoder(data->c->codec_id);
    if (!data->codec)
	return 0;
    if (avcodec_open(data->c, data->codec) < 0)
	return 0;

    data->outbuf = malloc(AVCODEC_MAX_AUDIO_FRAME_SIZE);
    data->seekpos = -1;
    data->player = 0;
    data->res = 1;
    data->limit = 0;
    obj->nr_channels = data->c->channels;
    
    memcpy(data->path, path, sizeof(data->path)-1);

    return 1;
}

static void wma_close(input_object *obj)
{
	struct wma_local_data *data;

	if (!obj)
		return;
	data = (struct wma_local_data *)obj->local_data;

	if (data)
	{
	    if(data->outbuf)
		free(data->outbuf);
	    avcodec_close(data->c);
	    av_close_input_file(data->ic);
	}
	
	if (obj->local_data) {
		free(obj->local_data);
		obj->local_data = NULL;
	}
}

static void wma_shutdown(void)
{
	return;
}

static input_plugin wma_plugin;

#ifdef __cplusplus
extern "C" {
#endif
	
input_plugin *input_plugin_info (void)
{
	memset(&wma_plugin, 0, sizeof(input_plugin));
	wma_plugin.version = INPUT_PLUGIN_VERSION;
	wma_plugin.name = "WMA player v0.1.1";
	wma_plugin.author = "Igor Mokrushin aka McMCC <mcmcc@mail.ru>";
	wma_plugin.init = wma_init;
	wma_plugin.shutdown = wma_shutdown;
	wma_plugin.can_handle = wma_can_handle;
	wma_plugin.open = wma_open;
	wma_plugin.close = wma_close;
	wma_plugin.play_frame = wma_play_frame;
	wma_plugin.frame_seek = wma_frame_seek;
	wma_plugin.frame_size = wma_frame_size;
	wma_plugin.nr_frames = wma_nr_frames;
	wma_plugin.frame_to_sec = wma_frame_to_sec;
	wma_plugin.sample_rate = wma_sample_rate;
	wma_plugin.channels = wma_channels;
	wma_plugin.stream_info = wma_stream_info;
	return &wma_plugin;
}

#ifdef __cplusplus
}
#endif
